import datetime
import pydantic


class Database(pydantic.BaseModel):
    user: str
    password: str
    host: str
    port: int
    echo: bool | None = None
    pool_size: int | None = None
    timeout: datetime.timedelta | None = None



import enum
import typing

import pydantic

###############
### General ###
###############


class TextLiteral(pydantic.BaseModel):
    charset: str | None
    text: str


class NullLiteral(pydantic.BaseModel):
    pass


Literal: typing.TypeAlias = TextLiteral | NullLiteral


class BinaryCharset(pydantic.BaseModel):
    pass


class Direction(enum.Enum):
    ASC = enum.auto()
    DESC = enum.auto()


#################
### Data Type ###
#################


class Tinyint(pydantic.BaseModel):
    unsigned: bool
    auto_increment: bool


class Smallint(pydantic.BaseModel):
    unsigned: bool
    auto_increment: bool


class Mediumint(pydantic.BaseModel):
    unsigned: bool
    auto_increment: bool


class Int(pydantic.BaseModel):
    unsigned: bool
    auto_increment: bool


class Bigint(pydantic.BaseModel):
    unsigned: bool
    auto_increment: bool


class Decimal(pydantic.BaseModel):
    precision: int | None
    scale: int | None


class Float(pydantic.BaseModel):
    pass


class Double(pydantic.BaseModel):
    pass


class Bit(pydantic.BaseModel):
    length: int | None


class Datetime(pydantic.BaseModel):
    precision: int | None
    default_now: bool
    on_update_now: bool


class Timestamp(pydantic.BaseModel):
    precision: int | None
    default_now: bool
    on_update_now: bool


class Time(pydantic.BaseModel):
    precision: int | None


class Date(pydantic.BaseModel):
    pass


class Year(pydantic.BaseModel):
    pass


class Char(pydantic.BaseModel):
    length: int
    charset: str | None
    collate: str | None


class Varchar(pydantic.BaseModel):
    length: int
    charset: str | None
    collate: str | None


class TinyText(pydantic.BaseModel):
    charset: str | None
    collate: str | None


class Text(pydantic.BaseModel):
    charset: str | None
    collate: str | None


class MediumText(pydantic.BaseModel):
    charset: str | None
    collate: str | None


class LongText(pydantic.BaseModel):
    charset: str | None
    collate: str | None


class Enum(pydantic.BaseModel):
    values: list[str]
    charset: str | None
    collate: str | None


class Set(pydantic.BaseModel):
    values: list[str]
    charset: str | None
    collate: str | None


class Binary(pydantic.BaseModel):
    length: int


class Varbinary(pydantic.BaseModel):
    length: int


class TinyBlob(pydantic.BaseModel):
    pass


class Blob(pydantic.BaseModel):
    pass


class MediumBlob(pydantic.BaseModel):
    pass


class LongBlob(pydantic.BaseModel):
    pass


class Geometry(pydantic.BaseModel):
    srid: int | None


class Point(pydantic.BaseModel):
    srid: int | None


class Linestring(pydantic.BaseModel):
    srid: int | None


class Polygon(pydantic.BaseModel):
    srid: int | None


class Geometrycollection(pydantic.BaseModel):
    srid: int | None


class Multipoint(pydantic.BaseModel):
    srid: int | None


class Multilinestring(pydantic.BaseModel):
    srid: int | None


class Multipolygon(pydantic.BaseModel):
    srid: int | None


class Json(pydantic.BaseModel):
    pass


IntegerDataType: typing.TypeAlias = Tinyint | Smallint | Mediumint | Int | Bigint
TextDataType: typing.TypeAlias = (
    Char | Varchar | TinyText | Text | MediumText | LongText | Enum | Set
)
SpatialDataType: typing.TypeAlias = (
    Geometry
    | Point
    | Linestring
    | Polygon
    | Geometrycollection
    | Multipoint
    | Multilinestring
    | Multipolygon
)

DataType: typing.TypeAlias = (
    IntegerDataType
    | Decimal
    | Float
    | Double
    | Bit
    | Datetime
    | Timestamp
    | Time
    | Date
    | Year
    | TextDataType
    | Binary
    | Varbinary
    | TinyBlob
    | Blob
    | MediumBlob
    | LongBlob
    | SpatialDataType
    | Json
)


##################
### Attributes ###
##################


class ColumnFormat(enum.Enum):
    FIXED = enum.auto()
    DYNAMIC = enum.auto()
    DEFAULT = enum.auto()


class StorageMedia(enum.Enum):
    DISK = enum.auto()
    MEMORY = enum.auto()
    DEFAULT = enum.auto()


class GenerationType(enum.Enum):
    VIRTUAL = enum.auto()
    STORED = enum.auto()


class NonNullableAttribute(pydantic.BaseModel):
    pass


class NullableAttribute(pydantic.BaseModel):
    pass


class LiteralDefaultAttribute(pydantic.BaseModel):
    value: Literal


class ExprDefaultAttribute(pydantic.BaseModel):
    expr: str


class NowDefaultAttribute(pydantic.BaseModel):
    pass


class GeneratedAttribute(pydantic.BaseModel):
    expr: str
    type: GenerationType


class OnUpdateNowAttribute(pydantic.BaseModel):
    pass


class CommentAttribute(pydantic.BaseModel):
    comment: TextLiteral


class InvisibleAttribute(pydantic.BaseModel):
    pass


class AutoIncrementAttribute(pydantic.BaseModel):
    pass


class UnsignedAttribute(pydantic.BaseModel):
    pass


class CharsetAttribute(pydantic.BaseModel):
    charset: str | BinaryCharset


class CollateAttribute(pydantic.BaseModel):
    collate: str


class SridAttribute(pydantic.BaseModel):
    srid: int


Attribute: typing.TypeAlias = (
    NonNullableAttribute
    | NullableAttribute
    | LiteralDefaultAttribute
    | ExprDefaultAttribute
    | NowDefaultAttribute
    | GeneratedAttribute
    | OnUpdateNowAttribute
    | CommentAttribute
    | ColumnFormat
    | StorageMedia
    | InvisibleAttribute
    | AutoIncrementAttribute
    | UnsignedAttribute
    | CharsetAttribute
    | CollateAttribute
    | SridAttribute
)

DefaultValue: typing.TypeAlias = ExprDefaultAttribute | LiteralDefaultAttribute


######################
### Create Options ###
######################


class Rowformat(enum.Enum):
    DEFAULT = enum.auto()
    DYNAMIC = enum.auto()
    FIXED = enum.auto()
    COMPRESSED = enum.auto()
    REDUNDANT = enum.auto()
    COMPACT = enum.auto()


class InsertMethod(enum.Enum):
    NO = enum.auto()
    FIRST = enum.auto()
    LAST = enum.auto()


class StorageType(enum.Enum):
    DISK = enum.auto()
    MEMORY = enum.auto()


class EngineCreateOption(pydantic.BaseModel):
    name: str


class SecondaryEngineCreateOption(pydantic.BaseModel):
    name: str


class AvgRowLength(pydantic.BaseModel):
    length: int


class AutoIncrementCreateOption(pydantic.BaseModel):
    value: int


class CommentCreateOption(pydantic.BaseModel):
    comment: TextLiteral


class StatsPersistentCreateOption(pydantic.BaseModel):
    option: bool


class RowFormatCreateOption(pydantic.BaseModel):
    format: Rowformat


class CharsetCreateOption(pydantic.BaseModel):
    charset: str | BinaryCharset


class CollateCreateOption(pydantic.BaseModel):
    collate: str


class TablespaceCreateOption(pydantic.BaseModel):
    name: str


CreateOption: typing.TypeAlias = (
    EngineCreateOption
    | SecondaryEngineCreateOption
    | AvgRowLength
    | AutoIncrementCreateOption
    | CommentCreateOption
    | StatsPersistentCreateOption
    | RowFormatCreateOption
    | CharsetCreateOption
    | CollateCreateOption
    | TablespaceCreateOption
)


class CreateOptions(pydantic.BaseModel):
    engine: str | None
    secondary_engine: str | None
    max_rows: int | None
    min_rows: int | None
    avg_row_length: int | None
    password: TextLiteral | None
    comment: str | None
    compression: TextLiteral | None
    encryption: TextLiteral | None
    auto_increment: int | None
    pack_keys: bool | None
    stats_auto_recalc: bool | None
    stats_persistent: bool | None
    stats_sample_pages: bool | None
    checksum: int | None
    delay_key_write: int | None
    row_format: Rowformat | None
    union: list[str] | None
    charset: str | None
    collate: str | None
    insert_method: InsertMethod | None
    data_directory: TextLiteral | None
    index_directory: TextLiteral | None
    tablespace: str | None
    storage: StorageType | None
    connection: TextLiteral | None
    key_block_size: int | None


##################
### Constraint ###
##################


class IdentifierKeyPart(pydantic.BaseModel):
    identifier: str
    length: int | None
    direction: Direction | None


KeyPart: typing.TypeAlias = IdentifierKeyPart


class IndexType(enum.Enum):
    BTREE = enum.auto()
    RTREE = enum.auto()
    HASH = enum.auto()


class ConstraintEnforcement(enum.Enum):
    ENFORCED = enum.auto()
    NOT_ENFORCED = enum.auto()


class IndexOptions(pydantic.BaseModel):
    key_block_size: int | None
    comment: str | None
    invisible: bool


class RefAction(enum.Enum):
    RESTRICT = enum.auto()
    CASCADE = enum.auto()
    SET_NULL = enum.auto()


class OnDeleteRefRule(pydantic.BaseModel):
    action: RefAction


class OnUpdateRefRule(pydantic.BaseModel):
    action: RefAction


RefRule: typing.TypeAlias = OnDeleteRefRule | OnUpdateRefRule


class References(pydantic.BaseModel):
    ref_table: str
    ref_columns: list[str]
    on_update: RefAction | None
    on_delete: RefAction | None


class PrimaryConstraint(pydantic.BaseModel):
    type: IndexType | None
    key_list: list[KeyPart]
    options: IndexOptions


class UniqueConstraint(pydantic.BaseModel):
    name: str | None
    type: IndexType | None
    key_list: list[KeyPart]
    options: IndexOptions


class IndexConstraint(pydantic.BaseModel):
    name: str | None
    type: IndexType | None
    key_list: list[KeyPart]
    options: IndexOptions


class FulltextConstraint(pydantic.BaseModel):
    name: str | None
    key_list: list[KeyPart]
    parser: str | None
    options: IndexOptions


class SpatialConstraint(pydantic.BaseModel):
    name: str | None
    key_list: list[KeyPart]
    options: IndexOptions


class ForeignConstraint(pydantic.BaseModel):
    name: str | None
    columns: list[str]
    references: References


class CheckConstraint(pydantic.BaseModel):
    name: str | None
    expr: str
    enforcement: ConstraintEnforcement


Constraint: typing.TypeAlias = (
    PrimaryConstraint
    | UniqueConstraint
    | IndexConstraint
    | FulltextConstraint
    | SpatialConstraint
    | ForeignConstraint
    | CheckConstraint
)


#################
### Partition ###
#################


class EnginePartitionOption_(pydantic.BaseModel):
    engine: str


type PartitionOption_ = EnginePartitionOption_


class PartitionOptions(pydantic.BaseModel):
    tablespace: str | None
    engine: str | None
    nodegroup: int | None
    max_rows: int | None
    min_rows: int | None
    data_directory: str | None
    index_directory: str | None
    comment: str | None


class KeyHashPartition(pydantic.BaseModel):
    name: str
    options: PartitionOptions


type ValuesLessThan = list[str | None]


class RangePartition(pydantic.BaseModel):
    name: str
    values: ValuesLessThan
    subpartitions: list[KeyHashPartition]
    options: PartitionOptions


type ValuesIn = list[str | list[str]]


class ListPartition(pydantic.BaseModel):
    name: str
    values: ValuesIn
    subpartitions: list[KeyHashPartition]
    options: PartitionOptions


class KeyPartitionType(pydantic.BaseModel):
    linear: bool
    algorithm: typing.Literal[1, 2] | None
    columns: list[str]


class HashPartitionType(pydantic.BaseModel):
    linear: bool
    expr: str


class Subpartitioning(pydantic.BaseModel):
    type: KeyPartitionType | HashPartitionType
    subpartition_count: int | None


class KeyHashPartitioning(pydantic.BaseModel):
    type: KeyPartitionType | HashPartitionType
    partition_count: int | None
    partitions: list[KeyHashPartition]
    sub: None = None


class RangePartitioning(pydantic.BaseModel):
    expr_or_columns: str | list[str]
    partitions: list[RangePartition]
    sub: Subpartitioning | None


class ListPartitioning(pydantic.BaseModel):
    expr_or_columns: str | list[str]
    partitions: list[ListPartition]
    sub: Subpartitioning | None


type Partitioning = KeyHashPartitioning | RangePartitioning | ListPartitioning


############
### Main ###
############


class Column(pydantic.BaseModel):
    name: str
    data_type: DataType

    non_nullable: bool | None
    default_value: DefaultValue | None
    comment: str | None
    invisible: bool

    format: ColumnFormat  # Default is ColumnFormat.DEFAULT
    storage_media: StorageMedia  # Default is StorageMedia.DEFAULT
    generated: GeneratedAttribute | None


class Table(pydantic.BaseModel):
    name: str
    columns: list[Column]
    constraints: list[Constraint]
    options: CreateOptions
    partitioning: Partitioning | None



from myrtille.mysql import types


def generate_ternary_option(option: bool):
    return '1' if option else '0'


def generate_ref_action(action: types.RefAction):
    match action:
        case types.RefAction.RESTRICT:
            return 'RESTRICT'
        case types.RefAction.CASCADE:
            return 'CASCADE'
        case types.RefAction.SET_NULL:
            return 'SET NULL'


def generate_key_part(key: types.KeyPart):
    res = f'`{key.identifier}`'
    if key.length is not None:
        res += f'({key.length})'
    if key.direction is not None:
        res += f' {key.direction.name}'
    return res


def generate_data_type(data_type: types.DataType):
    arguments: list[str] = []

    if (
        isinstance(data_type, types.Datetime | types.Timestamp | types.Time)
        and data_type.precision is not None
    ):
        arguments = [f'{data_type.precision}']

    if isinstance(
        data_type, types.Bit | types.Char | types.Varchar | types.Binary | types.Varbinary
    ):
        arguments = [f'{data_type.length}']

    if isinstance(data_type, types.Set | types.Enum):
        arguments = [f"'{v}'" for v in data_type.values]

    match data_type:
        case types.Tinyint():
            data_types_name = 'tinyint'
        case types.Smallint():
            data_types_name = 'smallint'
        case types.Mediumint():
            data_types_name = 'mediumint'
        case types.Int():
            data_types_name = 'int'
        case types.Bigint():
            data_types_name = 'bigint'
        case types.Decimal():
            data_types_name = 'decimal'
            if data_type.precision is not None:
                arguments = [f'{data_type.precision}', f'{data_type.scale or 0}']
        case types.Float():
            data_types_name = 'float'
        case types.Double():
            data_types_name = 'double'
        case types.Bit():
            data_types_name = 'bit'
        case types.Datetime():
            data_types_name = 'datetime'
        case types.Timestamp():
            data_types_name = 'timestamp'
        case types.Time():
            data_types_name = 'time'
        case types.Date():
            data_types_name = 'date'
        case types.Year():
            data_types_name = 'year'
        case types.Char():
            data_types_name = 'char'
        case types.Varchar():
            data_types_name = 'varchar'
        case types.TinyText():
            data_types_name = 'tinytext'
        case types.Text():
            data_types_name = 'text'
        case types.MediumText():
            data_types_name = 'mediumtext'
        case types.LongText():
            data_types_name = 'longtext'
        case types.Enum():
            data_types_name = 'enum'
        case types.Set():
            data_types_name = 'set'
        case types.Binary():
            data_types_name = 'binary'
        case types.Varbinary():
            data_types_name = 'varbinary'
        case types.TinyBlob():
            data_types_name = 'tinyblob'
        case types.Blob():
            data_types_name = 'blob'
        case types.MediumBlob():
            data_types_name = 'mediumblob'
        case types.LongBlob():
            data_types_name = 'longblob'
        case types.Json():
            data_types_name = 'json'
        case types.Geometry():
            data_types_name = 'geometry'
        case types.Point():
            data_types_name = 'point'
        case types.Linestring():
            data_types_name = 'linestring'
        case types.Polygon():
            data_types_name = 'polygon'
        case types.Geometrycollection():
            data_types_name = 'geomcollection'
        case types.Multipoint():
            data_types_name = 'multipoint'
        case types.Multilinestring():
            data_types_name = 'multilinestring'
        case types.Multipolygon():
            data_types_name = 'multipolygon'
    return data_types_name + (f'({",".join(arguments)})' if len(arguments) > 0 else '')


def generate_literal(literal: types.Literal):
    match literal:
        case types.TextLiteral():
            text = f"'{literal.text}'"
            return text
        case types.NullLiteral():
            return 'NULL'


def generate_default(default_value: types.DefaultValue):
    match default_value:
        case types.ExprDefaultAttribute():
            return f'({default_value.expr})'
        case types.LiteralDefaultAttribute():
            return generate_literal(default_value.value)


def generate_column(column: types.Column):
    attributes: list[str] = []

    if isinstance(column.data_type, types.IntegerDataType) and column.data_type.unsigned:
        attributes.append('unsigned')

    if isinstance(column.data_type, types.TextDataType):
        if column.data_type.charset is not None:
            attributes.append(f'CHARACTER SET {column.data_type.charset}')
        if column.data_type.collate is not None:
            attributes.append(f'COLLATE {column.data_type.collate}')

    if column.format != types.ColumnFormat.DEFAULT:
        attributes.append(f'/*!50606 COLUMN_FORMAT {column.format.name} */')

    if column.storage_media != types.StorageMedia.DEFAULT:
        attributes.append(f'/*!50606 STORAGE {column.storage_media.name} */')

    if column.generated is not None:
        attributes.append(
            f'GENERATED ALWAYS AS ({column.generated.expr}) {column.generated.type.name}'
        )

    if column.non_nullable is not None:
        attributes.append(('NOT ' if column.non_nullable else '') + 'NULL')

    if isinstance(column.data_type, types.SpatialDataType) and column.data_type.srid is not None:
        attributes.append(f'/*!80003 SRID {column.data_type.srid} */')

    if isinstance(column.data_type, types.IntegerDataType) and column.data_type.auto_increment:
        attributes.append('AUTO_INCREMENT')

    if column.default_value is not None:
        attributes.append(f'DEFAULT {generate_default(column.default_value)}')

    if isinstance(column.data_type, types.Datetime | types.Timestamp):
        if column.data_type.default_now:
            attributes.append('DEFAULT CURRENT_TIMESTAMP')
        if column.data_type.on_update_now:
            attributes.append('ON UPDATE CURRENT_TIMESTAMP')

    if column.comment is not None:
        attributes.append(f"COMMENT '{column.comment}'")
    if column.invisible:
        attributes.append('/*!80023 INVISIBLE */')

    return f'`{column.name}` {generate_data_type(column.data_type)}' + ''.join(
        f' {a}' for a in attributes
    )


def generate_constraint(constraint: types.Constraint):
    match constraint:
        case types.ForeignConstraint():
            col_names = f'({", ".join(f"`{c}`" for c in constraint.columns)})'
            ref_col_names = f'({", ".join(f"`{c}`" for c in constraint.references.ref_columns)})'
            parts = [
                f'CONSTRAINT `{constraint.name}` FOREIGN KEY {col_names}',
                f'REFERENCES `{constraint.references.ref_table}` {ref_col_names}',
            ]
            if constraint.references.on_delete is not None:
                parts.append(f'ON DELETE {generate_ref_action(constraint.references.on_delete)}')
            if constraint.references.on_update is not None:
                parts.append(f'ON UPDATE {generate_ref_action(constraint.references.on_update)}')
            return ' '.join(parts)
        case types.CheckConstraint():
            return f'CONSTRAINT `{constraint.name}` CHECK ({constraint.expr})'
        case _:
            key_list = f'({",".join(generate_key_part(key) for key in constraint.key_list)})'
            match constraint:
                case types.PrimaryConstraint():
                    return f'PRIMARY KEY {key_list}'
                case types.UniqueConstraint():
                    return f'UNIQUE KEY `{constraint.name}` {key_list}'
                case types.IndexConstraint():
                    return f'KEY `{constraint.name}` {key_list}'
                case types.FulltextConstraint():
                    return f'FULLTEXT KEY `{constraint.name}` {key_list}'
                case types.SpatialConstraint():
                    return f'SPATIAL KEY `{constraint.name}` {key_list}'


def generate_create_options(create_options: types.CreateOptions):
    attributes: list[str] = []
    if create_options.tablespace is not None:
        attributes.append(f'/*!50100 TABLESPACE `{create_options.tablespace}` */')
    if create_options.engine is not None:
        attributes.append(f'ENGINE={create_options.engine}')
    if create_options.auto_increment is not None:
        attributes.append(f'AUTO_INCREMENT={create_options.auto_increment}')
    if create_options.charset is not None:
        attributes.append(f'DEFAULT CHARSET={create_options.charset}')
    if create_options.collate is not None:
        attributes.append(f'COLLATE={create_options.collate}')
    if create_options.avg_row_length is not None:
        attributes.append(f'AVG_ROW_LENGTH={create_options.avg_row_length}')
    if create_options.stats_persistent is not None:
        attributes.append(
            f'STATS_PERSISTENT={generate_ternary_option(create_options.stats_persistent)}'
        )
    if create_options.row_format is not None:
        attributes.append(f'ROW_FORMAT={create_options.row_format.name}')
    if create_options.comment is not None:
        attributes.append(f"COMMENT='{create_options.comment}'")

    return ' '.join(attributes)


def generate_key_hash_partitioning_method(
    partitioning: types.KeyPartitionType | types.HashPartitionType,
):
    match partitioning:
        case types.KeyPartitionType():
            s = 'KEY'
        case types.HashPartitionType():
            s = f'HASH ({partitioning.expr})'
    return ('LINEAR ' if partitioning.linear else '') + s


def generate_partitioning_method(partitioning: types.Partitioning):
    match partitioning:
        case types.ListPartitioning() | types.RangePartitioning():
            match partitioning.expr_or_columns:
                case str():
                    args = f'({partitioning.expr_or_columns})'
                case list():
                    args = f'COLUMNS ({", ".join(partitioning.expr_or_columns)})'
            match partitioning:
                case types.RangePartitioning():
                    return f'RANGE {args}'
                case types.ListPartitioning():
                    return f'LIST {args}'
        case types.KeyHashPartitioning():
            return generate_key_hash_partitioning_method(partitioning.type)


def generate_partitioning(partitioning: types.Partitioning):
    return f'PARTITION BY {generate_partitioning_method(partitioning)}'


def generate_subpartitioning(subpartitioning: types.Subpartitioning):
    return f'SUBPARTITION BY {generate_key_hash_partitioning_method(subpartitioning.type)}'


def generate_partition_options(options: types.PartitionOptions):
    parts: dict[str, str] = {}
    if options.engine is not None:
        parts['ENGINE'] = options.engine
    return [f'{k} = {v}' for k, v in parts.items()]


def generate_key_hash_partition_def(partition: types.KeyHashPartition):
    return ' '.join((partition.name, *generate_partition_options(partition.options)))


def generate_partition_lt_values(values: types.ValuesLessThan):
    if values == [None]:
        args = 'MAXVALUE'
    else:
        args = f'({", ".join((v if v is not None else "MAXVALUE") for v in values)})'

    return f'VALUES LESS THAN {args}'


def generate_partition_in_values(values: types.ValuesIn):
    elements = [(f'({", ".join(v)})' if isinstance(v, list) else v) for v in values]
    return f'VALUES IN ({", ".join(elements)})'


def generate_subpartition(subparition: types.KeyHashPartition):
    return f'SUBPARTITION {generate_key_hash_partition_def(subparition)}'


def generate_partition(
    partition: types.RangePartition | types.ListPartition | types.KeyHashPartition,
):
    match partition:
        case types.RangePartition() | types.ListPartition():
            match partition:
                case types.RangePartition():
                    values = generate_partition_lt_values(partition.values)
                case types.ListPartition():
                    values = generate_partition_in_values(partition.values)
            parts = [partition.name, values, *generate_partition_options(partition.options)]
            definition = ' '.join(parts)

            subparitions = [generate_subpartition(sub) for sub in partition.subpartitions]
            if len(subparitions) > 0:
                definition += f'\n ({",\n  ".join(subparitions)})'
        case types.KeyHashPartition():
            definition = generate_key_hash_partition_def(partition)
    return f'PARTITION {definition}'


def generate_partitioning_clause(partitioning: types.Partitioning):
    parts = [generate_partitioning(partitioning)]
    if partitioning.sub is not None:
        parts.append(generate_subpartitioning(partitioning.sub))
    parts.append(f'({",\n ".join(generate_partition(p) for p in partitioning.partitions)})')
    return f'/*!50100 {"\n".join(parts)} */'


def generate(table: types.Table):
    elements = [generate_column(c) for c in table.columns] + [
        generate_constraint(c) for c in table.constraints
    ]
    options = generate_create_options(table.options)
    statement = (
        f'CREATE TABLE `{table.name}` (\n{",\n".join(f"  {e}" for e in elements)}\n) {options}'
    )
    if table.partitioning is not None:
        statement += '\n' + generate_partitioning_clause(table.partitioning)
    return statement



import contextlib
import time
import typing

import asyncmy
import pydantic

from myrtille.lib import cfg

ParamsType: typing.TypeAlias = typing.Collection[typing.Any]
Querriable: typing.TypeAlias = 'Database | Connection'


def _format_request(stmt: str, *, params: ParamsType | None = None):
    if params is not None:
        try:
            stmt = stmt % tuple(params)
        except Exception:
            if len(params) == 0:
                param_part = 'no params'
            else:
                param_part = f'params {", ".join(map(repr, params))}'
            stmt = f"Invalid stmt '{stmt}' with {param_part}"
    return f"Request '{' '.join(stmt.split())}'"


class Database(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    config: cfg.Database
    pool: asyncmy.pool.Pool = pydantic.Field(exclude=True)

    @contextlib.asynccontextmanager
    async def acquire(self):
        async with self.pool.acquire() as cnx:
            yield Connection(database=self, cnx=cnx)
            await cnx.rollback()

    async def execute(self, stmt: str, *, params: ParamsType | None = None):
        async with self.acquire() as cnx:
            await cnx.execute(stmt, params=params)
            await cnx.commit()

    async def fetch_all(self, stmt: str, *, params: typing.Sequence[typing.Any] | None = None):
        async with self.acquire() as cnx:
            return await cnx.fetch_all(stmt, params=params)

    async def fetch_optional(self, stmt: str, *, params: typing.Sequence[typing.Any] | None = None):
        async with self.acquire() as cnx:
            return await cnx.fetch_optional(stmt, params=params)


class Connection(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    database: Database
    cnx: asyncmy.connection.Connection

    async def execute(self, stmt: str, *, params: ParamsType | None = None):
        t0 = time.perf_counter()
        try:
            async with self.cnx.cursor() as cursor:
                await cursor.execute(stmt, params)
                if self.database.config.echo:
                    print(
                        f'Log: {_format_request(stmt, params=params)}: {time.perf_counter() - t0:2.2f}s'
                    )
        except Exception as e:
            e.add_note(f'In {_format_request(stmt, params=params)}')
            raise

    async def fetch_all(self, stmt: str, params: typing.Sequence[typing.Any] | None = None):
        t0 = time.perf_counter()
        try:
            async with self.cnx.cursor() as cursor:
                await cursor.execute(stmt, params)
                rows = await cursor.fetchall()
                if self.database.config.echo:
                    print(
                        f'Log: {_format_request(stmt, params=params)}: {time.perf_counter() - t0:2.2f}s'
                    )
                return rows
        except Exception as e:
            e.add_note(f'In {_format_request(stmt, params=params)}')
            raise

    async def fetch_optional(self, stmt: str, params: typing.Sequence[typing.Any] | None = None):
        rows = await self.fetch_all(stmt, params)
        if len(rows) > 1:
            raise Exception(
                f'{_format_request(stmt, params=params)} returned {len(rows)} (!= 1) rows'
            )
        elif len(rows) == 0:
            return None
        return rows[0]

    async def commit(self):
        await self.cnx.commit()


@contextlib.asynccontextmanager
async def make_database(db_config: cfg.Database):
    async with asyncmy.pool.create_pool(
        host=db_config.host,
        port=db_config.port,
        user=db_config.user,
        password=db_config.password,
        autocommit=False,
        echo=db_config.echo or False,
        connect_timeout=db_config.timeout.total_seconds()
        if db_config.timeout is not None
        else 31536000,
        minsize=db_config.pool_size or 1,
        maxsize=db_config.pool_size or 1,
    ) as pool:
        try:
            yield Database(config=db_config, pool=pool)
        finally:
            pool.close()
            await pool.wait_closed()



import asyncio
import re
import typing

import pydantic

from myrtille.lib import cfg, db


def _correct_ddl(ddl: str):
    # Removes display width on interer types
    # Removes floating point precision on time functions
    for s in [
        'tinyint',
        'smallint',
        'int',
        'bigint',
        'DEFAULT CURRENT_TIMESTAMP',
        'ON UPDATE CURRENT_TIMESTAMP',
    ]:
        ddl = re.sub(rf' {s}\([0-9]*\)', f' {s}', ddl, flags=re.IGNORECASE)

    # Removes non standart float precision
    for s in ['float', 'double']:
        ddl = re.sub(rf' {s}\([0-9]*,[0-9]*\)', f' {s}', ddl, flags=re.IGNORECASE)

    return ddl


class _Table(pydantic.BaseModel):
    schema_name: str
    table_name: str

    async def get_ddl(self, database: db.Database) -> str:
        show_create_response = await database.fetch_optional(
            f'SHOW CREATE TABLE `{self.schema_name}`.`{self.table_name}`'
        )
        assert show_create_response is not None
        (_, ddl) = show_create_response
        assert isinstance(ddl, str)
        return _correct_ddl(ddl)


async def export(schema_name: str, config: cfg.Database) -> typing.Sequence[str]:
    async with db.make_database(config) as database:
        return await asyncio.gather(
            *(
                _Table(schema_name=schema_name, table_name=table_name).get_ddl(database)
                for (table_name,) in await database.fetch_all(
                    """
                SELECT TABLE_NAME
                    FROM INFORMATION_SCHEMA.TABLES
                    WHERE TABLE_TYPE = "BASE TABLE"
                    AND TABLE_SCHEMA = %s
                """,
                    params=[schema_name],
                )
            )
        )



import dataclasses
import pathlib
import typing

import lark
import pydantic

from myrtille.mysql import types

GRAMMAR_FILE = pathlib.Path(__file__).parent / 'grammar.lark'
GRAMMAR_START = 'create_table'


Terminal: typing.TypeAlias = lark.lexer.Token
Meta: typing.TypeAlias = lark.tree.Meta


def v_args[R, **P](
    inline: bool = False, meta: bool = False
) -> typing.Callable[[typing.Callable[P, R]], typing.Callable[P, R]]:
    # type ignore because original lark.v_args typing does not work
    return lark.v_args(inline=inline, meta=meta, tree=False)  # type: ignore


def inline[R, **P](f: typing.Callable[P, R]) -> typing.Callable[P, R]:
    return v_args(inline=True)(f)


def get_child[R](children: typing.Sequence[R], *, position: int | None) -> R:
    if position is None:
        assert len(children) == 1
        return children[0]
    else:
        return children[position]


def through(position: int | None = None):
    def _handler(self: typing.Any, children: typing.Sequence[typing.Any]):
        return get_child(children, position=position)

    return _handler


def list_(self: typing.Any, children: list[typing.Any]):
    return children


def wrap_pydantic(model: type[pydantic.BaseModel], position: int | None = None):
    field_name = next(iter(model.model_fields))

    def _handler(self: typing.Any, children: typing.Sequence[typing.Any]):
        return model(**{field_name: get_child(children, position=position)})

    return _handler


def call(f: typing.Callable[[Terminal], typing.Any]):
    @inline
    def _handler(self: typing.Any, child: Terminal):
        return f(child)

    return _handler


def const(v: typing.Any):
    def _handler(self: typing.Any, children: typing.Any):
        return v

    return _handler


def const_call(f: typing.Callable[[], typing.Any]):
    def _handler(self: typing.Any, children: typing.Any):
        return f()

    return _handler


class Transformer(lark.Transformer[typing.Any, types.Table]):
    def __init__(self, original_text: str, visit_tokens: bool = True) -> None:
        self.original_text = original_text

        super().__init__(visit_tokens)

    ### Main

    @inline
    def column_definition(self, name: str, data_type: types.DataType, *attributes: types.Attribute):
        column = types.Column(
            name=name,
            data_type=data_type,
            non_nullable=None,
            default_value=None,
            comment=None,
            invisible=False,
            format=types.ColumnFormat.DEFAULT,
            storage_media=types.StorageMedia.DEFAULT,
            generated=None,
        )
        for attribute in attributes:
            match attribute:
                case types.NonNullableAttribute():
                    column.non_nullable = True
                case types.NullableAttribute():
                    column.non_nullable = False
                case types.LiteralDefaultAttribute():
                    column.default_value = attribute
                case types.ExprDefaultAttribute():
                    column.default_value = attribute
                case types.NowDefaultAttribute():
                    assert isinstance(column.data_type, (types.Datetime, types.Timestamp))
                    column.data_type.default_now = True
                case types.GeneratedAttribute():
                    column.generated = attribute
                case types.OnUpdateNowAttribute():
                    assert isinstance(column.data_type, (types.Datetime, types.Timestamp))
                    column.data_type.on_update_now = True
                case types.CommentAttribute():
                    column.comment = attribute.comment.text
                case types.ColumnFormat():
                    column.format = attribute
                case types.StorageMedia():
                    column.storage_media = attribute
                case types.InvisibleAttribute():
                    column.invisible = True
                case types.AutoIncrementAttribute():
                    assert isinstance(column.data_type, types.IntegerDataType)
                    column.data_type.auto_increment = True
                case types.UnsignedAttribute():
                    assert isinstance(column.data_type, types.IntegerDataType)
                    column.data_type.unsigned = True
                case types.CharsetAttribute():
                    assert isinstance(column.data_type, types.TextDataType)
                    if isinstance(attribute.charset, types.BinaryCharset):
                        column.data_type.charset = 'binary'
                    else:
                        column.data_type.charset = attribute.charset
                case types.CollateAttribute():
                    assert isinstance(column.data_type, types.TextDataType)
                    column.data_type.collate = attribute.collate
                case types.SridAttribute():
                    assert isinstance(column.data_type, types.SpatialDataType)
                    column.data_type.srid = attribute.srid
        return column

    table_element = through()
    table_element_list = list_

    @inline
    def create_table(
        self,
        name: str,
        elements: list[types.Column | types.Constraint],
        options: types.CreateOptions,
        partitioning: types.Partitioning | None,
    ):
        return types.Table(
            name=name,
            columns=[e for e in elements if isinstance(e, types.Column)],
            constraints=[e for e in elements if isinstance(e, types.Constraint)],
            options=options,
            partitioning=partitioning,
        )

    ### Types

    tinyint_data_type = const_call(lambda: types.Tinyint(unsigned=False, auto_increment=False))
    smallint_data_type = const_call(lambda: types.Smallint(unsigned=False, auto_increment=False))
    mediumint_data_type = const_call(lambda: types.Mediumint(unsigned=False, auto_increment=False))
    int_data_type = const_call(lambda: types.Int(unsigned=False, auto_increment=False))
    bigint_data_type = const_call(lambda: types.Bigint(unsigned=False, auto_increment=False))

    @inline
    def fixed_point_data_type(self, _: typing.Any, args: tuple[int, int]):
        return types.Decimal(precision=args[0], scale=args[1])

    float_data_type = const_call(types.Float)
    double_data_type = const_call(types.Double)

    @inline
    def bit_data_type(self, length: int | None):
        return types.Bit(length=length or 1)

    @inline
    def datetime_data_type(self, precision: int | None):
        return types.Datetime(precision=precision, default_now=False, on_update_now=False)

    @inline
    def timestamp_data_type(self, precision: int | None):
        return types.Timestamp(precision=precision, default_now=False, on_update_now=False)

    date_data_type = const_call(types.Date)

    @inline
    def time_data_type(self, precision: int | None):
        return types.Time(precision=precision)

    year_data_type = const_call(types.Year)

    @inline
    def char_data_type(self, _char: Terminal, length: int | None):
        return types.Char(length=length or 1, charset=None, collate=None)

    @inline
    def varchar_data_type(self, _char: Terminal, length: int):
        return types.Varchar(length=length, charset=None, collate=None)

    @inline
    def binary_data_type(self, _: Terminal, length: int | None):
        return types.Binary(length=length or 1)

    @inline
    def varbinary_data_type(self, length: int):
        return types.Varbinary(length=length)

    tinytext_data_type = const_call(lambda: types.TinyText(charset=None, collate=None))
    text_data_type = const_call(lambda: types.Text(charset=None, collate=None))
    mediumtext_data_type = const_call(lambda: types.MediumText(charset=None, collate=None))
    longtext_data_type = const_call(lambda: types.LongText(charset=None, collate=None))
    tinyblob_data_type = const_call(types.TinyBlob)
    blob_data_type = const_call(types.Blob)
    mediumblob_data_type = const_call(types.MediumBlob)
    longblob_data_type = const_call(types.LongBlob)

    @inline
    def enum_data_type(self, values: list[str]):
        return types.Enum(values=values, charset=None, collate=None)

    @inline
    def set_data_type(self, values: list[str]):
        return types.Set(values=values, charset=None, collate=None)

    geometry_data_type = const_call(lambda: types.Geometry(srid=None))
    point_data_type = const_call(lambda: types.Point(srid=None))
    linestring_data_type = const_call(lambda: types.Linestring(srid=None))
    polygon_data_type = const_call(lambda: types.Polygon(srid=None))
    geometrycollection_data_type = const_call(lambda: types.Geometrycollection(srid=None))
    multipoint_data_type = const_call(lambda: types.Multipoint(srid=None))
    multilinestring_data_type = const_call(lambda: types.Multilinestring(srid=None))
    multipolygon_data_type = const_call(lambda: types.Multipolygon(srid=None))

    json_data_type = const_call(types.Json)

    ### Attributes

    virtual_generation_type = const(types.GenerationType.VIRTUAL)
    stored_generation_type = const(types.GenerationType.STORED)

    non_nullable_attribute = const_call(types.NonNullableAttribute)
    nullable_attribute = const_call(types.NullableAttribute)
    literal_default_attribute = wrap_pydantic(types.LiteralDefaultAttribute)
    expr_default_attribute = wrap_pydantic(types.ExprDefaultAttribute)
    now_default_attribute = const_call(types.NowDefaultAttribute)

    @inline
    def generated_attribute(self, expr: str, type: types.GenerationType | None):
        return types.GeneratedAttribute(
            expr=expr, type=type if type is not None else types.GenerationType.VIRTUAL
        )

    on_update_attribute = const_call(types.OnUpdateNowAttribute)
    comment_attribute = wrap_pydantic(types.CommentAttribute)

    dynamic_column_format = const(types.ColumnFormat.DYNAMIC)
    fixed_column_format = const(types.ColumnFormat.FIXED)
    column_format_attribute = through()

    disk_storage_media = const(types.StorageMedia.DISK)
    memory_storage_media = const(types.StorageMedia.MEMORY)
    storage_attribute = through()

    auto_increment_attribute = const_call(types.AutoIncrementAttribute)
    unsigned_attribute = const_call(types.UnsignedAttribute)

    charset_attribute = wrap_pydantic(types.CharsetAttribute, position=1)
    collate_attribute = wrap_pydantic(types.CollateAttribute)

    custom_charset = through()
    binary_charset = const_call(types.BinaryCharset)

    invisible_attribute = const_call(types.InvisibleAttribute)
    srid_attribute = wrap_pydantic(types.SridAttribute)

    ### Constraints

    @inline
    def identifier_key_part(
        self, identifier: str, length: int | None, direction: types.Direction | None
    ):
        return types.IdentifierKeyPart(identifier=identifier, length=length, direction=direction)

    constraint_name = through()

    btree_index_type = const(types.IndexType.BTREE)
    rtree_index_type = const(types.IndexType.RTREE)
    hash_index_type = const(types.IndexType.HASH)
    index_type_clause = through()
    type_index_option = through()

    restrict_ref_action = const(types.RefAction.RESTRICT)
    cascade_ref_action = const(types.RefAction.CASCADE)
    set_null_ref_action = const(types.RefAction.SET_NULL)
    no_action_ref_action = const(None)

    on_delete_ref_rule = wrap_pydantic(types.OnDeleteRefRule)
    on_update_ref_rule = wrap_pydantic(types.OnUpdateRefRule)

    @inline
    def primary_constraint(
        self,
        _constraint_name: str | None,
        _name: str | None,
        index_type_clause: types.IndexType | None,
        key_list: list[types.KeyPart],
        *index_options: typing.Any,
    ):
        return types.PrimaryConstraint(
            type=index_type_clause,
            key_list=key_list,
            options=types.IndexOptions(key_block_size=None, comment=None, invisible=False),
        )

    @inline
    def unique_constraint(
        self,
        constraint_name: str | None,
        _key_or_index: typing.Any,
        name: str | None,
        index_type_clause: types.IndexType | None,
        key_list: list[types.KeyPart],
        *index_options: typing.Any,
    ):
        return types.UniqueConstraint(
            name=(name if name is not None else constraint_name),
            type=index_type_clause,
            key_list=key_list,
            options=types.IndexOptions(key_block_size=None, comment=None, invisible=False),
        )

    @inline
    def index_constraint(
        self,
        _key_or_index: typing.Any,
        name: str | None,
        index_type_clause: types.IndexType | None,
        key_list: list[types.KeyPart],
        *index_options: typing.Any,
    ):
        return types.IndexConstraint(
            name=name,
            type=index_type_clause,
            key_list=key_list,
            options=types.IndexOptions(key_block_size=None, comment=None, invisible=False),
        )

    @inline
    def fulltext_constraint(
        self,
        _key_or_index: typing.Any,
        name: str | None,
        key_list: list[types.KeyPart],
        *index_option: typing.Any,
    ):
        return types.FulltextConstraint(
            name=name,
            key_list=key_list,
            parser=None,
            options=types.IndexOptions(key_block_size=None, comment=None, invisible=False),
        )

    @inline
    def spatial_constraint(
        self,
        _key_or_index: typing.Any,
        name: str | None,
        key_list: list[types.KeyPart],
        *index_option: typing.Any,
    ):
        return types.SpatialConstraint(
            name=name,
            key_list=key_list,
            options=types.IndexOptions(key_block_size=None, comment=None, invisible=False),
        )

    @inline
    def references(
        self,
        ref_table: str,
        ref_columns: list[str] | None,
        _match: typing.Any,
        *ref_rules: types.RefRule,
    ):
        on_update = None
        on_delete = None
        for ref_rule in ref_rules:
            if isinstance(ref_rule, types.OnUpdateRefRule):
                on_update = ref_rule.action
            else:
                on_delete = ref_rule.action
        return types.References(
            ref_table=ref_table,
            ref_columns=ref_columns or [],
            on_update=on_update,
            on_delete=on_delete,
        )

    @inline
    def foreign_contraint(
        self,
        constraint_name: str | None,
        name: str | None,
        key_list: list[types.KeyPart],
        references: types.References,
    ):
        return types.ForeignConstraint(
            name=name if name is not None else constraint_name,
            columns=[k.identifier for k in key_list],
            references=references,
        )

    @inline
    def check_contraint(
        self,
        constraint_name: str | None,
        expr: str,
        enforcement: types.ConstraintEnforcement | None,
    ):
        return types.CheckConstraint(
            name=constraint_name,
            expr=expr,
            enforcement=enforcement
            if enforcement is not None
            else types.ConstraintEnforcement.ENFORCED,
        )

    ### Create Options

    zero_ternary_option = const(False)
    one_ternary_option = const(True)
    default_row_format = const(types.Rowformat.DEFAULT)
    dynamic_row_format = const(types.Rowformat.DYNAMIC)
    fixed_row_format = const(types.Rowformat.FIXED)
    compressed_row_format = const(types.Rowformat.COMPRESSED)
    redundant_row_format = const(types.Rowformat.REDUNDANT)
    compact_row_format = const(types.Rowformat.COMPACT)

    engine_create_option = wrap_pydantic(types.EngineCreateOption)
    secondary_engine_create_option = wrap_pydantic(types.SecondaryEngineCreateOption)
    avg_row_length_create_option = wrap_pydantic(types.AvgRowLength)
    comment_create_option = wrap_pydantic(types.CommentCreateOption)
    auto_increment_create_option = wrap_pydantic(types.AutoIncrementCreateOption)
    stats_persistent_create_option = wrap_pydantic(types.StatsPersistentCreateOption)
    row_format_create_option = wrap_pydantic(types.RowFormatCreateOption)
    charset_create_option = wrap_pydantic(types.CharsetCreateOption, position=1)
    collate_create_option = wrap_pydantic(types.CollateCreateOption)
    tablespace_create_option = wrap_pydantic(types.TablespaceCreateOption)

    def create_table_options(self, options: list[types.CreateOption]):
        create_options = types.CreateOptions(
            engine=None,
            secondary_engine=None,
            max_rows=None,
            min_rows=None,
            avg_row_length=None,
            password=None,
            comment=None,
            compression=None,
            encryption=None,
            auto_increment=None,
            pack_keys=None,
            stats_auto_recalc=None,
            stats_persistent=None,
            stats_sample_pages=None,
            checksum=None,
            delay_key_write=None,
            row_format=None,
            union=None,
            charset=None,
            collate=None,
            insert_method=None,
            data_directory=None,
            index_directory=None,
            tablespace=None,
            storage=None,
            connection=None,
            key_block_size=None,
        )
        for option in options:
            match option:
                case types.EngineCreateOption():
                    create_options.engine = option.name
                case types.SecondaryEngineCreateOption():
                    create_options.secondary_engine = option.name
                case types.AvgRowLength():
                    create_options.avg_row_length = option.length
                case types.AutoIncrementCreateOption():
                    create_options.auto_increment = option.value
                case types.CommentCreateOption():
                    create_options.comment = option.comment.text
                case types.StatsPersistentCreateOption():
                    create_options.stats_persistent = option.option
                case types.RowFormatCreateOption():
                    create_options.row_format = option.format
                case types.CharsetCreateOption():
                    if isinstance(option.charset, types.BinaryCharset):
                        create_options.charset = 'binary'
                    else:
                        create_options.charset = option.charset
                case types.CollateCreateOption():
                    create_options.collate = option.collate
                case types.TablespaceCreateOption():
                    create_options.tablespace = option.name

        return create_options

    ### Partition

    engine_partition_option = wrap_pydantic(types.EnginePartitionOption_)

    @inline
    def partition_options(self, *options_: types.PartitionOption_):
        options = types.PartitionOptions(
            tablespace=None,
            engine=None,
            nodegroup=None,
            max_rows=None,
            min_rows=None,
            data_directory=None,
            index_directory=None,
            comment=None,
        )
        for option in options_:
            match option:
                case types.EnginePartitionOption_():
                    options.engine = option.engine
        return options

    max_value = const(None)
    partition_lt_value = through()
    partition_lt_value_list = list_

    @inline
    def partition_values_lt(self, values: list[str | None] | None) -> types.ValuesLessThan:
        if values is None:
            values = [None]
        return values

    @inline
    def range_partition(
        self,
        name: str,
        values: types.ValuesLessThan,
        options: types.PartitionOptions,
        subpartitions: list[types.KeyHashPartition] | None,
    ):
        return types.RangePartition(
            name=name, values=values, subpartitions=subpartitions or [], options=options
        )

    range_partitions = list_

    @inline
    def key_hash_partition_def(self, name: str, options: types.PartitionOptions):
        return types.KeyHashPartition(name=name, options=options)

    key_hash_subpartition = through()
    key_hash_subpartitions = list_

    @inline
    def expr_or_columns(self, expr_or_columns: str | list[str] | None):
        if expr_or_columns is None:
            expr_or_columns = []
        return expr_or_columns

    @inline
    def hash_partition_type(self, linear: bool | None, expr: str):
        return types.HashPartitionType(linear=linear or False, expr=expr)

    @inline
    def subpartition_clause(self, type: types.HashPartitionType, count: int | None):
        return types.Subpartitioning(type=type, subpartition_count=count)

    @inline
    def range_partition_clause(
        self,
        expr_or_columns: str | list[str],
        subpartitioning: types.Subpartitioning | None,
        partitions: list[types.RangePartition],
    ):
        return types.RangePartitioning(
            expr_or_columns=expr_or_columns, partitions=partitions, sub=subpartitioning
        )

    partition_clause = through()

    ### General

    int_arg = through()

    @inline
    def int_pair_arg(self, i1: int, i2: int):
        return (i1, i2)

    key_list = list_
    text_list = list_

    desc_direction = const(types.Direction.DESC)
    asc_direction = const(types.Direction.ASC)

    ### Keywords

    charset = const(True)
    varchar = const(True)
    binary = const(True)
    key_or_index = const(True)

    ### Identifier

    unquoted_identifier = through()
    back_tick_text = call(lambda t: t[1:-1])
    identifier = through()
    dot_identifier = through()
    identifier_list = list_
    identifier_list_with_par = through()
    qual_identifier = const(True)

    ### Literal

    literal = through()

    int_literal = call(int)
    real_literal = call(float)

    text_literal = through(0)
    quoted_text = call(lambda t: t[1:-1])

    @inline
    def underscore_charset_text(self, underscore_charset: Terminal | None, text: str):
        return types.TextLiteral(
            charset=underscore_charset[1:] if underscore_charset is not None else None, text=text
        )

    null_literal = const_call(types.NullLiteral)

    # Expression

    expr_with_par = through()
    expr_list = list_

    @v_args(inline=False, meta=True)
    def expr(self, meta: Meta, children: typing.Any):
        return self.original_text[meta.start_pos : meta.end_pos]

    bool_pri = const(True)
    comp_op = const(True)
    predicate = const(True)
    predicate_operations = const(True)
    bit_expr = const(True)
    simple_expr = const(True)
    else_expression = const(True)
    then_expression = const(True)
    when_expression = const(True)
    array_cast = const(True)
    cast_type = const(True)
    charset_with_opt_binary = const(True)
    unicode = const(True)
    ascii = const(True)
    fulltext_options = const(True)
    ident_list_arg = const(True)
    ident_list = const(True)
    simple_identifier = const(True)
    query_expression_parens = const(True)
    locking_clause_list = const(True)
    locking_clause = const(True)
    locked_row_action = const(True)
    table_alias_ref_list = const(True)
    table_ref_with_wildcard = const(True)
    lock_strengh = const(True)
    query_expression = const(True)
    limit_clause = const(True)
    limit_options = const(True)
    limit_option = const(True)
    query_expression_body = const(True)
    union_option = const(True)
    query_primary = const(True)
    explicit_table = const(True)
    table_ref = const(True)
    table_value_constructor = const(True)
    row_value_explicit = const(True)
    values = const(True)
    query_specification = const(True)
    window_clause = const(True)
    window_definition = const(True)
    window_spec = const(True)
    having_clause = const(True)
    group_by_clause = const(True)
    olap_option = const(True)
    where_clause = const(True)
    from_clause = const(True)
    table_reference_list = const(True)
    table_reference = const(True)
    escaped_table_reference = const(True)
    joined_table = const(True)
    natural_join_type = const(True)
    outer_join_type = const(True)
    inner_join_type = const(True)
    table_factor = const(True)
    table_function = const(True)
    columns_clause = const(True)
    jt_column = const(True)
    on_empty_or_error = const(True)
    on_empty = const(True)
    on_error = const(True)
    jt_on_response = const(True)
    table_reference_list_parens = const(True)
    derived_table = const(True)
    single_table_parens = const(True)
    single_table = const(True)
    index_hint_list = const(True)
    index_hint = const(True)
    index_list = const(True)
    index_list_element = const(True)
    index_hint_clause = const(True)
    index_hint_type = const(True)
    table_alias = const(True)
    use_partition = const(True)
    into_clause = const(True)
    lines_clause = const(True)
    line_term = const(True)
    fields_clause = const(True)
    field_term = const(True)
    select_item_list = const(True)
    select_item = const(True)
    select_alias = const(True)
    table_wild = const(True)
    select_option = const(True)
    query_spec_option = const(True)
    with_clause = const(True)
    common_table_expression = const(True)
    window_function_call = const(True)
    null_treatment = const(True)
    lead_lag_info = const(True)
    simple_expr_with_parentheses = const(True)
    grouping_operation = const(True)
    sum_expr = const(True)
    json_function = const(True)
    windowing_clause = const(True)
    window_spec_details = const(True)
    window_frame_clause = const(True)
    window_frame_exclusion = const(True)
    window_frame_extent = const(True)
    window_frame_start = const(True)
    window_frame_between = const(True)
    window_frame_bound = const(True)
    window_frame_units = const(True)
    order_clause = const(True)
    order_list = const(True)
    order_expression = const(True)
    in_sum_expr = const(True)
    function_call = const(True)
    udf_expr_list = const(True)
    udf_expr = const(True)
    runtime_function_call = const(True)
    geometry_function = const(True)
    ws_num_codepoints = const(True)
    expr_list_with_parentheses = const(True)
    substring_function = const(True)
    date_time_ttype = const(True)
    interval = const(True)
    interval_time_stamp = const(True)
    trim_function = const(True)
    parentheses = const(True)
    json_operator = const(True)
    field_identifier = const(True)
    variable = const(True)
    system_variable = const(True)
    var_ident_type = const(True)
    user_variable = const(True)

    @inline
    def b_quote(self):
        pass

    @inline
    def bin_literal(self):
        pass

    @inline
    def bool_data_type(self):
        pass

    @inline
    def bool_literal(self):
        pass

    @inline
    def checksum_create_option(self):
        pass

    @inline
    def comment_index_option(self):
        pass

    @inline
    def comment_partition_option(self):
        pass

    @inline
    def compression_create_option(self):
        pass

    @inline
    def connection_create_option(self):
        pass

    @inline
    def data_directory_create_option(self):
        pass

    @inline
    def data_directory_partition_option(self):
        pass

    @inline
    def date_literal(self):
        pass

    @inline
    def default_column_format(self):
        pass

    @inline
    def default_storage_media(self):
        pass

    @inline
    def default_ternary_option(self):
        pass

    @inline
    def delay_key_write_create_option(self):
        pass

    @inline
    def encryption_create_option(self):
        pass

    @inline
    def enforced_constraint_enforcement(self):
        pass

    @inline
    def equal(self):
        pass

    @inline
    def expr_key_part(self):
        pass

    @inline
    def full_ref_match_type(self):
        pass

    @inline
    def hex_literal(self):
        pass

    @inline
    def index_directory_create_option(self):
        pass

    @inline
    def index_directory_partition_option(self):
        pass

    @inline
    def insert_method(self):
        pass

    @inline
    def insert_method_create_option(self):
        pass

    @inline
    def invisible_index_option(self):
        pass

    @inline
    def key_block_size_create_option(self):
        pass

    @inline
    def key_block_size_index_option(self):
        pass

    @inline
    def key_hash_partition(self):
        pass

    @inline
    def key_hash_partition_clause(self):
        pass

    @inline
    def key_hash_partitions(self):
        pass

    @inline
    def key_partition_type(self):
        pass

    @inline
    def list_partition(self):
        pass

    @inline
    def list_partition_clause(self):
        pass

    @inline
    def list_partitions(self):
        pass

    @inline
    def max_rows_create_option(self):
        pass

    @inline
    def max_rows_partition_option(self):
        pass

    @inline
    def min_rows_create_option(self):
        pass

    @inline
    def min_rows_partition_option(self):
        pass

    @inline
    def national_char_text(self):
        pass

    @inline
    def nchar(self):
        pass

    @inline
    def nchar_data_type(self):
        pass

    @inline
    def nodegroup_partition_option(self):
        pass

    @inline
    def not_enforced_constraint_enforcement(self):
        pass

    @inline
    def nvarchar(self):
        pass

    @inline
    def nvarchar_data_type(self):
        pass

    @inline
    def pack_keys_create_option(self):
        pass

    @inline
    def partial_ref_match_type(self):
        pass

    @inline
    def partition_in_value_list(self):
        pass

    @inline
    def partition_in_value_list_list(self):
        pass

    @inline
    def partition_values_in(self):
        pass

    @inline
    def password_create_option(self):
        pass

    @inline
    def qual_identifier_list(self):
        pass

    @inline
    def serial_data_type(self):
        pass

    @inline
    def simple_ref_match_type(self):
        pass

    @inline
    def stats_auto_recalc_create_option(self):
        pass

    @inline
    def stats_sample_pages_create_option(self):
        pass

    @inline
    def storage_create_option(self):
        pass

    @inline
    def storage_type(self):
        pass

    @inline
    def tablespace_partition_option(self):
        pass

    @inline
    def text_or_identifier(self, _: object):
        pass

    @inline
    def time_literal(self):
        pass

    @inline
    def timestamp_literal(self):
        pass

    @inline
    def union_create_option(self):
        pass

    @inline
    def visible_attribute(self):
        pass

    @inline
    def not_secondary_attribute(self):
        pass

    @inline
    def visible_index_option(self):
        pass

    @inline
    def with_parser_index_option(self):
        pass

    @inline
    def x_quote(self):
        pass

    @inline
    def zero_b(self):
        pass

    @inline
    def zero_x(self):
        pass

    @inline
    def zerofill_attribute(self):
        pass


@dataclasses.dataclass
class DDLParser:
    lark: lark.Lark

    @staticmethod
    def make():
        return DDLParser(
            lark=lark.Lark(
                GRAMMAR_FILE.read_text(encoding='utf-8'),
                start=GRAMMAR_START,
                debug=True,
                strict=True,
                maybe_placeholders=True,
                propagate_positions=True,
            )
        )

    def parse(self, ddl: str, transformer_cls: type[Transformer] | None = None):
        transformer = (transformer_cls or Transformer)(original_text=ddl)
        tree = self.lark.parse(ddl)
        return transformer.transform(tree)
